from fastapi import HTTPException, Security
from fastapi.security import APIKeyHeader

from metacatalog_api import core
from metacatalog_api import access_control

async def validate_api_key(api_key: str = Security(APIKeyHeader(name="X-API-Key"))):
    with core.connect() as session:
        token = access_control.validate_token(session, api_key) 
        if token is None:
            raise HTTPException(status_code=401, detail="Invalid API key")
        return token



from fastapi import Request, Depends
from starlette.middleware.cors import CORSMiddleware

from metacatalog_api.server import app, server

# these imports load the functionality needed for this metacatalog server
from metacatalog_api.apps.explorer.read import templates
from metacatalog_api.apps.explorer import static_files
from metacatalog_api.router.api.read import read_router as api_read_router
from metacatalog_api.router.api.create import create_router as api_create_router
from metacatalog_api.router.api.upload import upload_router
from metacatalog_api.apps.explorer.create import create_router as explorer_create
from metacatalog_api.apps.explorer.read import explorer_router
from metacatalog_api.apps.explorer.upload import upload_router as explorer_upload
from metacatalog_api.router.api.data import data_router
from metacatalog_api.router.api.security import validate_api_key

# at first we add the cors middleware to allow everyone to reach the API
app.add_middleware(
    CORSMiddleware, 
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"]
)

# the main page is defined here
# this can easily be changed to a different entrypoint
@app.get('/')
def index(request: Request):
    """
    Main page for the HTML Metacatalog Explorer.
    This example app includes the explorer read and create routes
    which are powered by the api route
    """
    return templates.TemplateResponse(request=request, name="index.html", context={"path": server.app_prefix, "root_path": server.uri_prefix})


# add all api routes - currently this is only splitted into read and create
app.include_router(api_read_router)
app.include_router(api_create_router, dependencies=[Depends(validate_api_key)])
app.include_router(upload_router, dependencies=[Depends(validate_api_key)])
app.include_router(data_router, dependencies=[Depends(validate_api_key)])

# add the default explorer application (the HTML)
app.mount(f"{server.app_prefix}static", static_files, name="static")
app.include_router(explorer_router, prefix=f"/{server.app_name}")
app.include_router(explorer_create, prefix=f"/{server.app_name}")
app.include_router(explorer_upload, prefix=f"/{server.app_name}")


if __name__ == '__main__':
    # run the server
    server.cli_cmd('default_server:app')



from contextlib import asynccontextmanager
import logging

from fastapi import FastAPI, Request
from pydantic_settings import BaseSettings, SettingsConfigDict
import uvicorn

from metacatalog_api import core
from metacatalog_api import __version__
from metacatalog_api.db import DB_VERSION


class Server(BaseSettings):
    model_config = SettingsConfigDict(
        cli_parse_args=True, 
        cli_prog_names="metacatalog-server"
    )
    host: str = "0.0.0.0"
    port: int = 8000
    root_path: str = ""
    reload: bool = False
    app_name: str = "explorer"
    autoupgrade: bool = False

    @property
    def uri_prefix(self):
        if self.root_path.startswith('/'):
            path = self.root_path
        else:
            path = f"/{self.root_path}"
        
        if not path.endswith('/'):
            path += '/'
        return path
    
    @property
    def app_prefix(self):
        path = self.uri_prefix
        
        if self.app_name.startswith('/'):
            path += self.app_name.strip('/')
        else:
            path += self.app_name
        
        if not path.endswith('/'):
            return f"{path}/"
        else:
            return path

    def cli_cmd(self, asgi_app: str):
        """Start the uvicorn server"""
        uvicorn.run(asgi_app, host=self.host, port=self.port, root_path=self.root_path, reload=self.reload)

logger = logging.getLogger('uvicorn.error')

# create the server object
server = Server()
logger.info(server.app_prefix, server.root_path, server.app_name)


# before we initialize the app, we check that the database is installed and up to date
@asynccontextmanager
async def lifespan(app: FastAPI):
    # check if the entries table can be found in the database
    with core.connect() as session:
        if not core.db.check_installed(session):
            logger.info("Database not installed, installing...")
            core.db.install(session, populate_defaults=True)
            logger.info("Database installed.")
    
    # after checking the database, we check the version
    with core.connect() as session:
            if core.db.has_version_mismatch(session):
                if server.autoupgrade:
                    core.migrate_db()
                else:
                    raise ValueError(f"Database version mismatch. Expected version {core.db.DB_VERSION}. Please run database migrations to update your schema.")

    # now we yield the application
    yield

    # here we can app tear down code - i.e. a log message

# build the base app
app = FastAPI(lifespan=lifespan)

@app.get('/version')
def get_version(request: Request):
    return {
        "metacatalog_api": __version__,
        "db_version": DB_VERSION,
        "hostname": request.url.hostname,
        "port": request.url.port,
        "root_path": request.url.path
    }

if __name__ == "__main__":
    print("The main server is not meant to be run directly. Check default_server.py for a sample application")



from pathlib import Path
import mimetypes

from fastapi import APIRouter
from fastapi.responses import StreamingResponse
from sqlmodel import text

from metacatalog_api import core


data_router = APIRouter()


def yield_error_message(error: str):
    yield f"""# Error
    The requested file could not be streamed back to you.
    I don't know why, but here is a error message for you:
    
    ```
    {error}
    ```
    """

def yield_file(file_path: str):
    with open(file_path, 'rb') as f:
        yield from f


def yield_internal_table(datasource):
    headers = []
    if datasource.temporal_scale is not None:
        headers.extend(datasource.temporal_scale.dimension_names)
    if datasource.spatial_scale is not None:
        headers.extend(datasource.spatial_scale.dimension_names)
    headers.extend(datasource.variable_names)
    yield ",".join(headers) + "\n" 
    sql = text(f"SELECT * FROM {datasource.path};")
    with core.connect() as session:
        for record in session.exec(sql):
            yield ",".join([str(c) for c in record]) + "\n"


@data_router.get('/entries/{entry_id}/dataset')
async def get_dataset(entry_id: int) -> StreamingResponse:
    # get the metadata
    entries = core.entries(ids=entry_id)
    if len(entries) == 0:
        return StreamingResponse(yield_error_message(f"Metadata Entry of id <ID={entry_id}> not found"), media_type="text/markdown")
    
    datasource = entries[0].datasource
    if datasource is None:
        return StreamingResponse(yield_error_message(f"Metadata Entry of id <ID={entry_id}> has no datasource"), media_type="text/markdown")

    # These are two cases that are not yet supported:
    if '*' in datasource.path:
        return StreamingResponse(yield_error_message(f"MetaCatalog API does currently not support streaming of wildcard paths"), media_type="text/markdown")
    elif Path(datasource.path).is_dir(): 
        return StreamingResponse(yield_error_message(f"Metadata Entry of id <ID={entry_id}> points to a directory. GZip result streaming is not yet supported."), media_type="text/markdown")
    
    
    if datasource.type.name == "internal":
        return StreamingResponse(yield_internal_table(datasource), media_type="text/csv")
    elif datasource.type.name == "external":
        return StreamingResponse(yield_error_message(f"Metadata Entry of id <ID={entry_id}> is external and cannot be downloaded"), media_type="text/markdown")
    elif datasource.type.name == "csv":
        return StreamingResponse(yield_file(datasource.path), media_type="text/csv")
    elif datasource.type.name == "local":
        mime, _ = mimetypes.guess_type(datasource.path)
        if mime is None:
            mime = "application/octet-stream"
        return StreamingResponse(yield_file(datasource.path), media_type=mime)
    elif datasource.type.name == "netCDF":
        return StreamingResponse(yield_file(datasource.path), media_type="application/netcdf")
    else:
        return StreamingResponse(yield_error_message(f"Metadata Entry of id <ID={entry_id}> has an unknown datasource type"), media_type="text/markdown")



import hashlib
from datetime import datetime
from sqlmodel import SQLModel, Field, Relationship, Session, select
from random import choice
from string import ascii_letters, digits

from metacatalog_api.models import PersonTable, Author


class UserTokenBase(SQLModel):
    token_hash: str = Field(min_length=64, max_length=64)
    created_at: datetime = Field(default_factory=datetime.now)
    valid_until: datetime | None = None

class UserTokenTable(UserTokenBase, table=True):
    __tablename__ = 'user_access_tokens'

    id: int = Field(primary_key=True)
    user_id: int | None = Field(default=None,foreign_key='persons.id')
    user: PersonTable | None = Relationship()


class UserToken(UserTokenBase):
    user: Author |  None = None


def register_new_token(session: Session, user: Author | None, valid_until: datetime | None = None) -> str:
    #new_key = Fernet.generate_key().decode()
    new_key = 'k' + ''.join(choice(ascii_letters + digits) for i in range(31))
    token_hash = hashlib.sha256(new_key.encode('utf-8')).hexdigest()

    token = UserTokenTable(
        user_id=user.id if user is not None else None,
        token_hash=token_hash,
        valid_until=valid_until
    )

    session.add(token)
    session.commit()
    return new_key 


def validate_token(session: Session, token: str) -> UserToken | None:
    token_hash = hashlib.sha256(token.encode('utf-8')).hexdigest()
    user_token = session.exec(select(UserTokenTable).where(UserTokenTable.token_hash ==token_hash)).first()

    return user_token



from typing import List, Generator
import os
from pathlib import Path
from contextlib import contextmanager
from datetime import datetime

from sqlmodel import Session, create_engine, text
from metacatalog_api import models
from dotenv import load_dotenv
from pydantic_geojson import FeatureCollectionModel

from metacatalog_api import db
from metacatalog_api.file_uploads import UploadCache
from metacatalog_api import access_control


load_dotenv()

METACATALOG_URI = os.environ.get("METACATALOG_URI", 'postgresql://metacatalog:metacatalog@localhost:5432/metacatalog')
SQL_DIR = Path(__file__).parent / "sql"

cache = UploadCache()


@contextmanager
def connect(url: str = None) -> Generator[Session, None, None]:
    uri = url if url is not None else METACATALOG_URI
    engine = create_engine(uri)

    with Session(engine) as session:
        yield session


def get_session(url: str = None) -> Session:
    if url is None:
        url = os.getenv('METACATALOG_URI')
        
    engine = create_engine(url)
    return Session(engine)


def migrate_db(schema: str = 'public') -> None:
    # get the current version
    with connect() as session:
        current_version = db.get_db_version(session, schema=schema)['db_version']
    
    # as long as the local _DB_VERSION is higher than the remote version, we can load a migration
    if current_version < db.DB_VERSION:
        # get the migration sql
        migration_sql = db.load_sql(SQL_DIR / 'migrate' / f'migration_{current_version + 1}.sql').format(schema=schema)
        
        # run
        with connect() as session:
            session.exec(text(migration_sql))
            
            # update the db version
            session.exec(text(f"INSERT INTO {schema}.metacatalog_info (db_version) VALUES ({current_version + 1});")) 
            session.commit()
        # inform the user
        print(f"Migrated database from version {current_version} to {current_version + 1}")
        
        # finally call the migration function recursively
        migrate_db()


def register_token(user: models.Author | None = None, valid_until: datetime | None = None):
    with connect() as session:
        new_key = access_control.register_new_token(session, user, valid_until)

    print(f"Generated a new token. Save this token in a save space as it will not be displayed again:\n{new_key}\n")


def entries(offset: int = 0, limit: int = None, ids: int | List[int] = None, full_text: bool = True, search: str = None, variable: str | int = None, title: str = None, geolocation: str = None) -> list[models.Metadata]:
    # check if we filter or search
    with connect() as session:
        if search is not None:
            search_results = db.search_entries(session, search, limit=limit, offset=offset, variable=variable, full_text=full_text, geolocation=geolocation)

            if len(search_results) == 0:
                return []
            # in any other case get them by id
            # in any other case, request the entries by id
            results = db.get_entries_by_id(session=session, entry_ids=[r.id for r in search_results])

            return results
        elif ids is not None:
            results = db.get_entries_by_id(session, ids, limit=limit, offset=offset)
        else:
            results = db.get_entries(session, limit=limit, offset=offset, variable=variable, title=title, geolocation=geolocation)

    return results


def entries_locations(ids: int | List[int] = None, limit: int = None, offset: int = None, search: str = None, filter: dict = {}) -> FeatureCollectionModel:
    # handle the ids
    if ids is None:
        ids = []
    if isinstance(ids, int):
        ids = [ids]
    
    # check if we filter or search
    with connect() as session:
        # run the search to ge the ids
        if search is not None:
            search_results = db.search_entries(session, search, limit=limit, offset=offset)
            ids = [*ids, *[r.id for r in search_results]]
        
            # if no ids have been found, return an empty FeatureCollection
            if len(ids) == 0:
                return {"type": "FeatureCollection", "features": []}
        
        # in any other case we go for the locations.
        result = db.get_entries_locations(session, ids=ids, limit=limit, offset=offset)
    
    return result


def groups(id: int = None, title: str = None, description: str = None, type: str = None, entry_id: int = None, with_metadata: bool = False, limit: int = None, offset: int = None):
    with connect() as session:
        if id is not None or (title is not None and '%' not in title):
            group = db.get_group(session, id=id, title=title, with_metadata=with_metadata)
            return group
        else:
            groups = db.get_groups(session, title=title, description=description, type=type, entry_id=entry_id, limit=limit, offset=offset)
            return groups 


def group_types():
    with connect() as session:
        return db.get_grouptypes(session)


def licenses(id: int = None, offset: int = None, limit: int = None) -> models.License | list[models.License]:
    with connect() as session:
        if id is not None:
            result = db.get_license_by_id(session, id=id)
        else:
            result = db.get_licenses(session, limit=limit, offset=offset)
    
    return result


def authors(id: int = None, entry_id: int = None, search: str = None, name: str = None, exclude_ids: List[int] = None, offset: int = None, limit: int = None) -> List[models.Author]:
    with connect() as session:
        # if an author_id is given, we return only the author of that id
        if id is not None:
            authors = db.get_author_by_id(session, id=id)
        # if an entry_id is given, we return only the authors of that entry
        elif entry_id is not None:
            authors = db.get_authors_by_entry(session, entry_id=entry_id)
        elif name is not None:
            authors = db.get_authors_by_name(session, name=name, limit=limit, offset=offset)
        else:
            authors = db.get_authors(session, search=search, exclude_ids=exclude_ids, limit=limit, offset=offset)
    
    return authors


def author(id: int = None, name: str = None, search: str = None) -> models.Author | None:
    with connect() as session:
        if id is not None:
            author = db.get_author_by_id(session, id=id)
        elif name is not None:
            author = db.get_author_by_name(session, name=name)
        else:
            authors = db.get_authors(session, search=search, limit=1) 
            if len(authors) == 0:
                author = None
            else:
                author = authors[0]
    return author


def variables(id: int = None, only_available: bool = False, offset: int = None, limit: int = None) -> List[models.Variable]:
    with connect() as session:
        if only_available:
            variables = db.get_available_variables(session, limit=limit, offset=offset)
        elif id is not None:
            variables = db.get_variable_by_id(session, id=id)
        else:
            variables = db.get_variables(session, limit=limit, offset=offset)
    
    return variables


def datatypes(id: int = None) -> List[models.DatasourceTypeBase]:
    # TODO: this may need some more parameters
    with connect() as session:
        return db.get_datatypes(session, id=id)


def add_author(payload: models.AuthorCreate, no_duplicates: bool = True) -> models.Author:
    with connect() as session:
        if no_duplicates:
            author = db.create_or_get_author(session, payload)
        else:
            author = db.add_author(session, payload)
    
    return author


def add_entry(payload: models.EntryCreate, author_duplicates: bool = False) -> models.Metadata:
    # add the entry
    with connect() as session:
        entry = db.add_entry(session, payload=payload, author_duplicates=author_duplicates)
    
        # check if there was a datasource
        if payload.datasource is not None:
            # if the path is in the UploadCache, the file was already uploaded and just needs to be copied
            if payload.datasource.path in cache:
                new_path = cache.save_to_data(file_hash=payload.path)
                payload.datasource.path = new_path

            entry = db.add_datasource(session, entry_id=entry.id, datasource=payload.datasource)
        session.commit()

        # handle groups
        if payload.groups is not None and len(payload.groups) > 0:
            for group in payload.groups:
                db.group_entries(session=session, group=group, entry_ids=[entry.id])
    return entry


def add_datasource(entry_id: int, payload: models.DatasourceCreate) -> models.Metadata:
    # if the path is in the UploadCache, the file was already uploaded and just needs to be copied
    if payload.path in cache:
        new_path = cache.save_to_data(file_hash=payload.path)
        payload.path = new_path

    with connect() as session:
        entry = db.add_datasource(session, entry_id=entry_id, datasource=payload)

    return entry


def add_group(payload: models.EntryGroupCreate) -> models.EntryGroup:
    with connect() as session:
        group = db.add_group(
            session=session,
            title=payload.title,
            description=payload.description, 
            type=payload.type,
            entry_ids=payload.entry_ids
        )
    return group



import os

from metacatalog_api import core


def migrate(schema: str = 'public'):
    """
    Migrate the database schema to the latest version.
    :param schema: The schema holding the tables, usually 'public' (default)
    """
    core.migrate_db(schema=schema)


if __name__ == '__main__':
    import fire
    fire.Fire(migrate)
    


from pathlib import Path
from fastapi import APIRouter, Request
from fastapi.templating import Jinja2Templates

from metacatalog_api import core
from metacatalog_api.server import server

create_router = APIRouter()
templates = Jinja2Templates(directory=Path(__file__).parent / 'templates')


@create_router.get('/page/create_metadata.html')
def new_entry_page(request: Request):
    return templates.TemplateResponse(request=request, name="page_create_metadata.html", context={"path": server.app_prefix})

@create_router.get('/create/entries.html')
def new_entry(request: Request):
    return templates.TemplateResponse(request=request, name="add_entry.html", context={"path": server.app_prefix})


@create_router.get("/utils/leaflet_draw.html")
def leaflet_draw(request: Request, geom: str = 'marker'):
    if geom.lower() == 'marker':
        return templates.TemplateResponse(request=request, name="leaflet_marker.html", context={})
    elif geom.lower() == 'extent':
        return templates.TemplateResponse(request=request, name="leaflet_extent.html", context={})


@create_router.get('/create/authors.html')
def new_author(request: Request):
    return templates.TemplateResponse(request=request, name="author.html", context={"path": server.app_prefix})


@create_router.get('/create/details.html')
def new_details(request: Request):
    return templates.TemplateResponse(request=request, name="details.html", context={"path": server.app_prefix})


@create_router.get('/create/datasources.html')
def new_datasource(request: Request):
    # load the datasource types
    types = core.datatypes()
    return templates.TemplateResponse(request=request, name="add_datasource.html", context={"types": types, "path": server.app_prefix})



from fastapi import APIRouter
from fastapi import UploadFile

from metacatalog_api.core import cache


upload_router = APIRouter()

@upload_router.post('/uploads')
def create_new_upload_preview(file: UploadFile, guess_metadata: bool = False):
    file_hash = cache.index_file(file)

    file_info = cache.get_file(file_hash)
    return {
        'file_hash': file_hash,
        'filename': file_info.filename,
        'size': file_info.size
    }

@upload_router.get('/uploads')
def get_all_upload_previews():
    file_infos = list(cache.cache.values())

    return {
        'count': len(file_infos),
        'files': file_infos
    }

